//////////////////////////////////////////////////////////////////////
// Author: Jim MacNair
// Version: 1.0
// Date: 09/01/98
//
// Description:
//
// This program implements a multi-threaded in-process COM automation
// server which allows programs which support COM to use the CICS
// client External Presentation Interface (EPI).
//
// This program is written for the Windows NT environment and 
// should also work with Windows 95 and Windows 98.
//
//////////////////////////////////////////////////////////////////////

// api3270.cpp : implementation file
//

static char Version[]=
"@(#)EPI3270 V1.00 - 3270 EPI Subroutines - Jim MacNair ("__DATE__")";

#include "stdafx.h"
#include "comdef.h"
#include "largeint.h"
#include "epi3270.h"
#include "api3270.h"
#include "epimsg.h"

////////////////////////////////////////////////////////////////
//                                                            //
// Defines and global variables                               //
//                                                            //
////////////////////////////////////////////////////////////////

#define ATTR_CHAR  '@'          // default screen dump attribute char 
#define UNPROT_CHAR  '~'        // default screen dump unprot attr char 
#define MAX_LINE	100			// Maximum characters in error line
#define FIELD_LEN   41          // Maximum characters in field def line

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// api3270

IMPLEMENT_DYNCREATE(api3270, CCmdTarget)

api3270::api3270()
{
	EnableAutomation();
	
	// To keep the application running as long as an OLE automation 
	//	object is active, the constructor calls AfxOleLockApp.

	// Create an instance of a 3270 screen object
	screen = new(C3270);

	// Initialize the terminal index
	TermIndex=CICS_EPI_TERM_INDEX_NONE;

	// Initialize the terminal data areas
	InitializeTerminal();

	AfxOleLockApp();
}

api3270::~api3270()
{
	// Write out a trace record
	screen->CICS_3270_WriteTraceLine("api3270 Destructor entered\n", CICS_3270_TRACE_MIN);

	// Delete the terminal, if one is still around
	if (TermIndex != CICS_EPI_TERM_INDEX_NONE)
	{
		DeleteTerminal();
	}

	// Close the trace
	screen->CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_OFF, "");

	// Delete our screen object
	delete screen;

	// To terminate the application when all objects created with
	// 	with OLE automation, the destructor calls AfxOleUnlockApp.
	
	AfxOleUnlockApp();
}


void api3270::OnFinalRelease()
{
	// When the last reference for an automation object is released
	// OnFinalRelease is called.  The base class will automatically
	// deletes the object.  Add additional cleanup required for your
	// object before calling the base class.

	// Try to uninstall the CICS terminal, ignoring any errors
	DeleteTerminal();

	CCmdTarget::OnFinalRelease();
}


BEGIN_MESSAGE_MAP(api3270, CCmdTarget)
	//{{AFX_MSG_MAP(api3270)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

BEGIN_DISPATCH_MAP(api3270, CCmdTarget)
	//{{AFX_DISPATCH_MAP(api3270)
	DISP_PROPERTY(api3270, "nextransid", m_nextransid, VT_BSTR)
	DISP_PROPERTY(api3270, "numberFields", m_numberFields, VT_I2)
	DISP_PROPERTY(api3270, "transid", m_transid, VT_BSTR)
	DISP_PROPERTY(api3270, "numberUnprotFields", m_numberUnprotFields, VT_I2)
	DISP_PROPERTY(api3270, "rows", m_rows, VT_I2)
	DISP_PROPERTY(api3270, "columns", m_columns, VT_I2)
	DISP_PROPERTY(api3270, "termname", m_termname, VT_BSTR)
	DISP_PROPERTY(api3270, "lastevent", m_lastevent, VT_I2)
	DISP_PROPERTY(api3270, "tranCount", m_tranCount, VT_I4)
	DISP_PROPERTY(api3270, "endTranCount", m_endTranCount, VT_I4)
	DISP_PROPERTY(api3270, "atiCount", m_atiCount, VT_I4)
	DISP_PROPERTY(api3270, "convCount", m_convCount, VT_I4)
	DISP_PROPERTY(api3270, "sendCount", m_sendCount, VT_I4)
	DISP_PROPERTY(api3270, "lastResponse", m_lastResponse, VT_I4)
	DISP_PROPERTY(api3270, "bytesSent", m_bytesSent, VT_I4)
	DISP_PROPERTY(api3270, "bytesReceived", m_bytesReceived, VT_I4)
	DISP_PROPERTY(api3270, "readBufferCount", m_readBufferCount, VT_I4)
	DISP_PROPERTY(api3270, "totalResponses", m_totalResponses, VT_I4)
	DISP_PROPERTY(api3270, "totalResponseTime", m_totalResponseTime, VT_I4)
	DISP_FUNCTION(api3270, "BeginTran", BeginTran, VT_I2, VTS_BSTR VTS_BSTR)
	DISP_FUNCTION(api3270, "SendInput", SendInput, VT_I2, VTS_BSTR)
	DISP_FUNCTION(api3270, "WaitATI", WaitATI, VT_I2, VTS_PVARIANT)
	DISP_FUNCTION(api3270, "ModifyField", ModifyField, VT_I2, VTS_I2 VTS_I2 VTS_BSTR)
	DISP_FUNCTION(api3270, "ModifyFieldNum", ModifyFieldNum, VT_I2, VTS_I2 VTS_BSTR)
	DISP_FUNCTION(api3270, "SetCursor", SetCursor, VT_I2, VTS_I2 VTS_I2)
	DISP_FUNCTION(api3270, "ClearScreen", ClearScreen, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(api3270, "GetFieldData", GetFieldData, VT_I2, VTS_I2 VTS_I2 VTS_PVARIANT)
	DISP_FUNCTION(api3270, "GetFieldDataNum", GetFieldDataNum, VT_I2, VTS_I2 VTS_PVARIANT)
	DISP_FUNCTION(api3270, "GetFieldAttribute", GetFieldAttribute, VT_I2, VTS_I2 VTS_I2 VTS_PVARIANT VTS_PVARIANT VTS_PVARIANT)
	DISP_FUNCTION(api3270, "GetFieldRowCol", GetFieldRowCol, VT_I2, VTS_I2 VTS_PVARIANT VTS_PVARIANT)
	DISP_FUNCTION(api3270, "GetFieldLength", GetFieldLength, VT_I2, VTS_I2 VTS_I2 VTS_PVARIANT)
	DISP_FUNCTION(api3270, "keyin", keyin, VT_I2, VTS_BSTR)
	DISP_FUNCTION(api3270, "CreateTerminal", CreateTerminal, VT_I2, VTS_BSTR VTS_BSTR VTS_BSTR)
	DISP_FUNCTION(api3270, "WaitEvent", WaitEvent, VT_I2, VTS_PVARIANT)
	DISP_FUNCTION(api3270, "CheckEvent", CheckEvent, VT_I2, VTS_PVARIANT)
	DISP_FUNCTION(api3270, "GetScreenSum", GetScreenSum, VT_I2, VTS_NONE)
	DISP_FUNCTION(api3270, "EPIErrorMessage", EPIErrorMessage, VT_BSTR, VTS_I2)
	DISP_FUNCTION(api3270, "GetDataAtLoc", GetDataAtLoc, VT_I2, VTS_I2 VTS_I2 VTS_I2 VTS_PVARIANT)
	DISP_FUNCTION(api3270, "DeleteTerminal", DeleteTerminal, VT_I2, VTS_NONE)
	DISP_FUNCTION(api3270, "SetTrace", SetTrace, VT_I2, VTS_I2 VTS_BSTR)
	DISP_FUNCTION(api3270, "DumpFields", DumpFields, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(api3270, "WriteFields", WriteFields, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(api3270, "WriteTraceLine", WriteTraceLine, VT_EMPTY, VTS_BSTR VTS_I2)
	DISP_FUNCTION(api3270, "TraceScreen", TraceScreen, VT_EMPTY, VTS_BSTR VTS_BSTR)
	DISP_FUNCTION(api3270, "GetCursorLoc", GetCursorLoc, VT_I2, VTS_NONE)
	DISP_FUNCTION(api3270, "CheckATI", CheckATI, VT_I2, VTS_PVARIANT)
	DISP_FUNCTION(api3270, "GetScreenImage", GetScreenImage, VT_BSTR, VTS_BSTR VTS_BSTR)
	DISP_FUNCTION(api3270, "GetFields", GetFields, VT_BSTR, VTS_NONE)
	DISP_FUNCTION(api3270, "GetProtTable", GetProtTable, VT_BSTR, VTS_NONE)
	DISP_FUNCTION(api3270, "GetEventTrace", GetEventTrace, VT_BSTR, VTS_NONE)
	DISP_FUNCTION(api3270, "GetDataByNum", GetDataByNum, VT_BSTR, VTS_I2)
	DISP_FUNCTION(api3270, "ServerCount", ServerCount, VT_I2, VTS_NONE)
	DISP_FUNCTION(api3270, "ServerName", ServerName, VT_BSTR, VTS_I2)
	DISP_FUNCTION(api3270, "SetTraceLevel", SetTraceLevel, VT_EMPTY, VTS_I2)
	DISP_FUNCTION(api3270, "GetUnprotFieldRowCol", GetUnprotFieldRowCol, VT_I2, VTS_I2 VTS_PVARIANT VTS_PVARIANT)
	DISP_FUNCTION(api3270, "ModifyUnprotFieldNum", ModifyUnprotFieldNum, VT_I2, VTS_I2 VTS_BSTR)
	//}}AFX_DISPATCH_MAP
END_DISPATCH_MAP()

// Note: we add support for IID_Iapi3270 to support typesafe binding
//  from VBA.  This IID must match the GUID that is attached to the 
//  dispinterface in the .ODL file.

// {AFCB2CB3-1A98-11D2-BC8F-204C4F4F5020}
static const IID IID_Iapi3270 =
{ 0xafcb2cb3, 0x1a98, 0x11d2, { 0xbc, 0x8f, 0x20, 0x4c, 0x4f, 0x4f, 0x50, 0x20 } };

BEGIN_INTERFACE_MAP(api3270, CCmdTarget)
	INTERFACE_PART(api3270, IID_Iapi3270, Dispatch)
END_INTERFACE_MAP()

// {AFCB2CB4-1A98-11D2-BC8F-204C4F4F5020}
IMPLEMENT_OLECREATE(api3270, "epi3270.api3270", 0xafcb2cb4, 0x1a98, 0x11d2, 0xbc, 0x8f, 0x20, 0x4c, 0x4f, 0x4f, 0x50, 0x20)

/////////////////////////////////////////////////////////////////////////////
// api3270 message handlers

/////////////////////////////////////////////////////////////////////////////
// Routine to initialize the terminal performance and statistics 
/////////////////////////////////////////////////////////////////////////////

void api3270::InitializeTerminal()

{
	int	i;

	// Initialize the instance variables
	m_numberFields = 0;
	m_numberUnprotFields = 0;
	m_rows=24;
	m_columns=80;
	m_tranCount = 0;
	m_endTranCount = 0;
	m_atiCount = 0;
	m_sendCount = 0;
	m_convCount = 0;
	m_readBufferCount = 0;
	m_bytesSent = 0;
	m_bytesReceived = 0;
	m_lastResponse = 0;
	m_totalResponses = 0;
	m_totalResponseTime = 0;
	m_termname = _T("");
	m_transid = _T("");
	m_nextransid = _T("");

	for (i=0; i < MAX_SAVED_EVENTS; i++)
	{
		// Clear the event table and event data
		lastEvents[i] = 0;
		lastEventData[4*i]     = ' ';
		lastEventData[4*i + 1] = ' ';
		lastEventData[4*i + 2] = ' ';
		lastEventData[4*i + 3] = ' ';
		lastEventTime[i] = 0;
	}
}

/////////////////////////////////////////////////////////////////////////////
// Routine to execute a CICS transaction
// This routine will get its input data from the screen buffer.
// It will issue a start transaction and will then wait for
// the transaction to end or for an input request.
// It will then return to the calling program, returning 
// a 0 if the transaction has ended or the transaction
// has requested more input data.  It will return an EPI error
// message if one is received.  The calling program can determine
// if a transaction is running by examining the m_transid property.
/////////////////////////////////////////////////////////////////////////////

short api3270::BeginTran(LPCTSTR transid, LPCTSTR AIDType) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	rc;
	int		length;
	int		transcount;		// work variable
	char	aid;
	char*	inptdata;
	char*	transdata;
	char	trans[CICS_EPI_TRANSID_MAX + 1];
	char	errmsg[CICS_3270_MAX_EPI_MSG_LEN];
	char	errline[MAX_LINE + 50];
	SYSTEMTIME		curtime;
	FILETIME		filtime;

	// Write a trace record
	sprintf(errline,"\nBeginTran Entered - Transid = %s, AID Key = %s\n", transid, AIDType);
	screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

	// check if transaction is already running
	if (!m_transid.IsEmpty())
	{
		// Write an error message
		sprintf(errline,"Transaction already running - %s\n",m_transid);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
		screen->CICS_3270_Write_Error_Log(errline);
		return 99;
	}

	// Get the transaction id passed in
	memcpy(trans, transid, sizeof(trans));

	// Process the aid byte into a value
	aid = parseAid(AIDType);

	// Build the inbound data stream from the screen buffer
	screen->CICS_3270_GetInbound(aid, &inptdata, &length);

	// Check if the transaction code is null
	if (trans[0] == '\0')
	{
		// Try to get a transaction code from the inbound data stream
		// skip past the aid byte and cursor address
		transdata = inptdata + 3;
		// check for an SBA sequence
		if (transdata[0] == CICS_3270_ORDER_SBA)
		{
			// Skip past the SBA sequence
			transdata += 3;
		}

		// clear the trans field
		memset(trans, '\0', sizeof(trans));

		transcount = 0;
		while ((transdata[0] > ' ') && (transcount < CICS_EPI_TRANSID_MAX))
		{
			// get the next character of the transaction id
			trans[transcount] = transdata[transcount];

			// move to the next character
			transcount++;
		}
	}

	if (trans[0] != '\0')
	{
      rc = CICS_EpiStartTran(TermIndex,
							(char *) trans,
							(cics_ubyte_t *) inptdata,
							(cics_ushort_t) length);

		if (rc == CICS_EPI_NORMAL)
		{
			// Increment the transaction count
			m_tranCount++;

			// Remember when the transaction started
			// Get the current time
			GetSystemTime(&curtime);
			SystemTimeToFileTime(&curtime, &filtime);

			// Save the time in the work variable
			timestart.HighPart = filtime.dwHighDateTime;
			timestart.LowPart = filtime.dwLowDateTime;

			// Remember the number of bytes sent
			m_bytesSent += length;

			// Set the transaction id variable
			m_transid = trans;
			m_nextransid.Empty();

			// write a trace record
			sprintf(errline, "Transaction (%s) started\n", trans);
			screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
		}
		else
		{
			// Write a trace record
			sprintf(errline, "Transaction (%s) failed to start\n", trans);
			screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// Log the error 
			screen->CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
			sprintf(errline, "***Error in CICS_EpiStartTran - Return code %d - %s\n", rc,errmsg);
			screen->CICS_3270_Write_Error_Log(errline);
			screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
		}
	}
	else
	{
		// set dummy return code to indicate error 
		rc = 100;

		// Log the error 
		sprintf(errline, "Transid missing - nothing started\n");
		screen->CICS_3270_Write_Error_Log(errline);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
	} // endif 

	// release the storage that was acquired by the screen->CICS_3270_GetInbound routine
	free(inptdata);

	return rc;
}

short api3270::SendInput(LPCTSTR AIDType) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char	aid;         // 3270 aid byte   
	short	rc;          // EPI return code 
	char	*data;       // Inbound Data Stream    
	int		maxchar=0;   // Characters in input datastream 
	int		length;      // Data Length            
	char	errmsg[CICS_3270_MAX_EPI_MSG_LEN];
	char	errline[MAX_LINE + 50];
	SYSTEMTIME		curtime;
	FILETIME		filtime;

	// Send the input to CICS

	// Write a trace record
	sprintf(errline, "\nSendInput entered for attention key %s\n", AIDType);
	screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

	// Process the aid byte
	aid = parseAid(AIDType);

	// Build the 3270 data stream to send in,   
	// from the current 3270 screen image.      
	screen->CICS_3270_GetInbound(aid,&data,&length);

	// Send the data to CICS
	rc = CICS_EpiReply(TermIndex,
                       (u_byte_t *) data,
                       (u_short_t) length);

	if (rc == CICS_EPI_NORMAL)
	{
		// Remember when the reply was sent
		// Get the current time
		GetSystemTime(&curtime);
		SystemTimeToFileTime(&curtime, &filtime);

		// Save the time in the work variable
		timestart.HighPart = filtime.dwHighDateTime;
		timestart.LowPart = filtime.dwLowDateTime;

		// Remember number of byt4es sent
		m_bytesSent += length;
	}

	// Free the storage for the input data stream
	free(data);

	// Write trace entry
	screen->CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
	sprintf(errline, "%d: CICS_EpiReply %s with aid of %s\n",
			rc,errmsg,AIDType);
	screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

	// Check for errors
	if (rc != CICS_EPI_NORMAL)
	{
		// Log any errors
		screen->CICS_3270_Write_Error_Log(errline);
	}

	return rc;
}

short api3270::ModifyField(short row, short col, LPCTSTR fieldData) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short		rc=0;			// return code
	char		errline[MAX_LINE + 50];
	CICS_3270_Field_t	*field;	// field characteristics

	//
	// Get the field referenced by the (row,column)
	//
	screen->CICS_3270_GetField(row, col, &field);

	// Check if there is a field here
	if (field->row == -1)
	{
		// Error - indicate no such field
		sprintf(errline, "01 ModifyField - no such field - (%d,%d)\n", row, col);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to the caller with error
		rc = 1;
	}
	else
	{
		// Make sure the field is unprotected
		if ((field->attribute & CICS_3270_ATTR_PROT) > 0)
		{
			// Error - the field is protected
			sprintf(errline, "02 ModifyField - field is protected - (%d,%d)\n", row, col);

			// log the error
			screen->CICS_3270_Write_Error_Log(errline);

			// write a trace record
			screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// return to the caller with error
			rc = 2;
		}
		else
		{
			// Modify the field
			rc = screen->CICS_3270_ModifyField(row, col, fieldData);

			if (rc == 0)
			{
				// build a trace entry
				sprintf(errline, "ModifyField - field modified - (%d,%d) - %s\n", 
						row, col, fieldData);

				// write a trace record
				screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
			}
		}
	}

	return rc;
}

short api3270::modifybynum(short fieldNum, LPCTSTR fieldData, int fldtype, int numflds)

{
	short		row;
	short		col;
	short		rc=0;
	char		errline[MAX_LINE + 50];

	// Preset the error message, in case we need it
	// Error - invalid field number
	// indicate the error
	sprintf(errline, "*Error* 01 modifybynum - no such field - %d\n", fieldNum);

	// Check for invalid field number
	if ((fieldNum < 1) || (fieldNum > m_numberFields))
	{
		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// Indicate bad field number
		rc = 3;
	}
	else
	{
		// Get the row and column attributes
		screen->CICS_3270_GetFieldLocByNum(fieldNum, &row, &col, fldtype);

		if (row == -1)
		{
			// log the error
			screen->CICS_3270_Write_Error_Log(errline);

			// write a trace record
			screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// return to the caller with error
			rc = 1;
		}
		else
		{
			// Modify the field
			rc = screen->CICS_3270_ModifyField(row, col, fieldData);

			// Check if it worked
			if (rc == 0)
			{
				// Write a trace record
				sprintf(errline, "modifybynum - field# - %d, data = %s\n", fieldNum, fieldData);
				screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
			}
		}
	}

	return rc;
}

// Routine to modify a field in the 3270 screen image, using 
// a field number to indicate the field.  If the field number
// is invalid, a return code of 1 is returned.  All fields
// are counted.  The first field is field 1.
// The screen is searched starting from the upper left corner,
// with the first attribute byte found counting as the first
// field.  If a field wraps around the screen, then this field
// will be counted as the last field on the screen.

short api3270::ModifyFieldNum(short fieldNum, LPCTSTR fieldData) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	return modifybynum(fieldNum, fieldData, CICS_3270_FIELDS_ALL, m_numberFields);
}

// Routine to modify a field in the 3270 screen image, using 
// a field number to indicate the field.  If the field number
// is invalid, a return code of 1 is returned.  Only unprotected
// fields are counted.  The first unprotected field is field 1.
// The screen is searched starting from the upper left corner,
// with the first attribute byte found counting as the first
// field.  If a field wraps around the screen, then this field
// will be counted as the last field on the screen.

short api3270::ModifyUnprotFieldNum(short fieldNum, LPCTSTR fieldData) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	return modifybynum(fieldNum, fieldData, CICS_3270_FIELDS_UNPROT, m_numberUnprotFields);
}

short api3270::GetCursorLoc() 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	// return the current position of the cursor
	return screen->cursorPosition;
}

short api3270::SetCursor(short row, short col) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	rc = 0;
	char	errline[MAX_LINE + 50];

	if ((row < 0) || (row >= m_rows) || (col < 0) || (col >= m_columns))
	{
		// Indicate the error
		sprintf(errline, "1 - SetCursor - Invalid row (%d) or column (%d)\n", row, col);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
		screen->CICS_3270_Write_Error_Log(errline);

		rc = 1;
	}
	else
	{
		// write a trace record
		sprintf(errline, "SetCursor to row (%d) column (%d)\n", row, col);

		// set the cursor location
		screen->CICS_3270_SetCursor(row, col);
	}

	return rc;
}

void api3270::ClearScreen() 
{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char    errline[MAX_LINE + 50];

	// Clear the Buffer
	screen->CICS_3270_ClearBuffer();

	// Reset the instance variables
	m_numberFields = (short) screen->CICS_3270_getNumberFields();
	m_numberUnprotFields = (short) screen->CICS_3270_getNumberUnprotFields();
	screen->CICS_3270_GetScreenParms(&m_rows, &m_columns);


	// Write a trace record
	sprintf(errline, "\nClearScreen entered - clearing screen - transid = %s nextran = %s\n",
			m_transid, m_nextransid);
	screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

	if ((m_transid.GetLength() > 0) || (m_nextransid.GetLength() > 0))
	{
		// Send the clear to CICS
		if (m_transid.GetLength() > 0)
		{
			// Write a trace record
			sprintf(errline,"ClearScreen - sending clear to CICS - transaction is %s\n",m_transid);
			screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// we have an active transaction - send the clear as input
			SendInput("CLEAR");
		}
		else
		{
			// Write a trace record
			sprintf(errline,"ClearScreen - sending clear to CICS - next trans is %s\n",m_nextransid);
			screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// we must use the next transaction id to send the input
			BeginTran(m_nextransid, "CLEAR");
		}
	}
}

////////////////////////////////////////////////////
//
// Routine to get a complete screen image and return
// it in a string.  Attribute bytes will be displayed
// as an "@" character.  Characters less than a space
// will be displayed as a blank.
//
////////////////////////////////////////////////////

BSTR api3270::GetScreenImage(LPCTSTR protchar, LPCTSTR unprotchar) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char	buff[43*134 + 1];
	CString	tdata;
	char prot;
	char unprot;

	// Get the substitution characters
	prot = protchar[0];
	unprot = unprotchar[0];

	// take defaults if protchar or unprotchar is null
	if (prot < ' ')
	{
		prot = ATTR_CHAR;
	}

	if (unprot < ' ')
	{
		unprot = UNPROT_CHAR;
	}

	int i;
	i = sizeof(buff);

	// clear the screen buffer area 
	memset(buff, '\0', sizeof(buff));

	// build the screen buffer 
	screen->CICS_3270_GetScreenImage(buff, prot, unprot);

	// Transfer it to a CString
	tdata = buff;

	return tdata.AllocSysString();
}

////////////////////////////////////////////////////
//
// This routine will return a single string 
// with a fixed length substring for each field
// on the screen.  The substring will contain the
// characteristics of each field.
//
////////////////////////////////////////////////////

BSTR api3270::GetFields() 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char	buff[1000*FIELD_LEN + 1];	// Allow for 1000 fields
	CString	tdata;

	// clear the screen buffer area 
	memset(buff, '\0', sizeof(buff));

	// Get the field definitions 
	screen->CICS_3270_GetFieldAttrs(buff);

	// Transfer it to a CString
	tdata = buff;

	return tdata.AllocSysString();
}

////////////////////////////////////////////////////
//
// This routine will return a single string 
// with a single character for each position on the
// screen.  The character will be a '0' for unprotected
// fields, a '1' for protected fields, a '2' for
// unprotected attributes and a '3' for protected 
// attribute bytes.
//
////////////////////////////////////////////////////

BSTR api3270::GetProtTable() 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char	buff[27*132 + 1];	// Allow for max size screen
	CString	tdata;

	// Get the field definitions 
	screen->CICS_3270_GetProtTable(buff);

	// Transfer it to a CString
	tdata = buff;

	// Return the string to the calling program
	return tdata.AllocSysString();
}

////////////////////////////////////////////////////
//
// This routine will access the data in a field
// at a given screen location and return the data
// as a string.  The row and column parameters must
// point to a valid attribute byte.
//
////////////////////////////////////////////////////

short api3270::GetFieldData(short row, short col, VARIANT FAR* varFieldData) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	rc=0;	// return code
	int		maxchars;
	CICS_3270_Field_t	*field;	// field characteristics
	wchar_t* unidata;	// Unicode characters
	char    errline[MAX_LINE + 50];

	//
	// Get the field referenced by the (row,column)
	//
	screen->CICS_3270_GetField(row, col, &field);

	// Check if there is a field here
	if (field->row == -1)
	{
		// Error - indicate no such field
		sprintf(errline, "02 GetFieldData - no such field - (%d,%d)\n", row, col);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to the caller with error
		rc = 2;
	}
	else
	{
		// get the length of the field data
		maxchars = 2*strlen(field->data) + 2;

		// Get seme temporary storage for the Unicode data
		unidata = (wchar_t *) malloc(maxchars);

		// Initialize the result
		unidata[0] = 0;

		// copy the data into a unicode string
		int worked = MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,field->data,-1,unidata,maxchars);

		// write a trace record
		sprintf(errline, "GetFieldData - (%d,%d) - data - %.80s\n", row, col,field->data);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// Clear the passed in Variant, releasing any current data
		VariantClear(varFieldData);

		// Set the variant type to string
		V_VT(varFieldData) = VT_BSTR;

		// Set the string value
		V_BSTR(varFieldData) = ::SysAllocString(unidata);

		// Release the temporary storage
		free(unidata);
	}

	// release the data areas
	screen->CICS_3270_FreeField(field);

	return rc;
}

////////////////////////////////////////////////////
//
// This routine will get the contents of a 3270
// field.  The field is identified by the field
// number, with the first field on the screen
// defined as field 1.  All fields are counted,
// and data from protected as well as unprotected
// fields an be accessed.  
//
////////////////////////////////////////////////////

short api3270::GetFieldDataNum(short fieldNum, VARIANT FAR* varFieldData) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	row;
	short	col;
	short	rc=0;
	char    errline[MAX_LINE + 50];

	// Get the row and column attributes
	screen->CICS_3270_GetFieldLocByNum(fieldNum, &row, &col, CICS_3270_FIELDS_ALL);

	if (row == -1)
	{
		// Error - invalid field number
		// indicate the error
		sprintf(errline, "03 GetFieldDataNum - no such field - %d\n", fieldNum);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to the caller with error
		rc = 3;
	}
	else
	{
		// Write a trace entry
		sprintf(errline, "GetFieldDataNum - field# - %d\n", fieldNum);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// Get the field data
		GetFieldData(row, col, varFieldData);
	}

	return rc;
}

////////////////////////////////////////////////////
//
// This routine will access the data in a given
// field identified by the field number.  The data 
// is returned as a string.  If the field number
// is invalid, a null string will be returned.
//
////////////////////////////////////////////////////

BSTR api3270::GetDataByNum(short fieldNum) 

{
	CString strResult;
	// TODO: Add your dispatch handler code here
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	// Initialize the return variable
	strResult = "";

	short	row;
	short	col;
	short	rc=0;
	char    errline[MAX_LINE + 50];
	CICS_3270_Field_t	*field;	// field characteristics

	// Get the row and column attributes
	screen->CICS_3270_GetFieldLocByNum(fieldNum, &row, &col, CICS_3270_FIELDS_ALL);

	if (row == -1)
	{
		// Error - invalid field number
		// indicate the error
		sprintf(errline, "03 GetDataByNum - no such field - %d\n", fieldNum);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to the caller with error
		rc = 3;
	}
	else
	{
		//
		// Get the field referenced by the (row,column)
		//
		screen->CICS_3270_GetField(row, col, &field);

		// Write a trace entry
		sprintf(errline, "GetDataByNum - field# - %d length %d\n", fieldNum, strlen(field->data));
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// copy the field data into a CString
		strResult = field->data;

		// release the field data areas
		screen->CICS_3270_FreeField(field);
	}

	return strResult.AllocSysString();
}

/////////////////////////////////////////////////////////////////////
//
// Routine to get screen data starting at a particular row and
// column.
//
/////////////////////////////////////////////////////////////////////

short api3270::GetDataAtLoc(short row, short col, short numchars, VARIANT FAR* varData) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	rc=0;
	int		bufferOffset=0;
	char	scrdata[8193];
	wchar_t* unidata;	// Unicode characters
	char    errline[MAX_LINE + 50];

	// Initialize our work variables
	scrdata[0] = '\0';

	if ((row < 0) || (row >= m_rows) || (col < 0) || (col >= m_columns) || (numchars <= 0))
	{
		// bad row or column or maxlength <= 0
		rc = 1;

		// Write a trace and error log record
		sprintf(errline,"GetDataAtLoc - Invalid row (%d) or column (%d) or length (%d)\n", row, col, numchars);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
		screen->CICS_3270_Write_Error_Log(errline);
	}
	else
	{
		// calculate the buffer location from the row and column
		bufferOffset = row*m_columns + col;

		// Make sure we don't overflow our internal buffer
		if (numchars > sizeof(scrdata) - 1)
		{
			numchars = sizeof(scrdata) - 1;
		}

		// Get the data
		screen->CICS_3270_GetDataAtLoc(bufferOffset, numchars, scrdata);

		// Write a trace record, with a maximum of 80 characters of data
		sprintf(errline,"\nGetDataAtLoc %d bytes at (%d,%d) - data = %.80s\n",
				strlen(scrdata), row, col, scrdata);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// Get storage for a temporary Unicode buffer
		unidata = (wchar_t *) malloc(2*numchars + 2);

		// Initialize the result variant to be a string
		unidata[0] = 0;

		// copy the data into a unicode string
		int worked = MultiByteToWideChar(CP_ACP,MB_PRECOMPOSED,scrdata,-1,unidata,33*133);

		// Clear the passed in Variant, releasing any current data
		VariantClear(varData);

		// Set the variant type to string
		V_VT(varData) = VT_BSTR;

		// Set the string value
		V_BSTR(varData) = ::SysAllocString(unidata);

		// Release the temporary storage
		free(unidata);
	}

	return rc;
}

short api3270::GetFieldAttribute(short row, short col, VARIANT FAR* attr, VARIANT FAR* extattr, VARIANT FAR* color) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int		rc=0;	// return code
	CICS_3270_Field_t	*field;	// field characteristics
	char    errline[MAX_LINE + 50];

	//
	// Get the field referenced by the (row,column)
	//
	screen->CICS_3270_GetField(row, col, &field);

	// Check if there is a field here
	if (field->row == -1)
	{
		// Error - indicate no such field
		sprintf(errline, "02 GetFieldData - no such field - (%d,%d)\n", row, col);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to the caller with error
		rc = 2;
	}
	else
	{
		// Clear the current values of the return variables
		VariantClear(attr);
		VariantClear(extattr);
		VariantClear(color);
		// Set the variant data types to short
		V_VT(attr) = VT_I2;
		V_VT(extattr) = VT_I2;
		V_VT(color) = VT_I2;

		// Set the return data in the variants
		V_I2(attr) = field->attribute;
		V_I2(extattr) = field->extattr;
		V_I2(color) = field->color;
	}

	// release the data areas
	screen->CICS_3270_FreeField(field);

	return rc;
}

short api3270::GetUnprotFieldRowCol(short fieldNum, VARIANT FAR* row, VARIANT FAR* col) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int		rc=0;
	char    errline[MAX_LINE + 50];
	short	temprow;
	short	tempcol;

	// Get the row and column attributes
	screen->CICS_3270_GetFieldLocByNum(fieldNum, &temprow, &tempcol, CICS_3270_FIELDS_UNPROT);

	if (temprow == -1)
	{
		// Error - invalid field number
		// indicate the error
		sprintf(errline, "03 GetUnprotFieldRowCol - no such field - %d\n", fieldNum);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to the caller with error
		rc = 3;
	}
 	else
	{
		// Clear the current data values
		VariantClear(row);
		VariantClear(col);

		// Set the variant data types to short
		V_VT(row) = VT_I2;
		V_VT(col) = VT_I2;

		// Set the data values in the variants
		V_I2(row) = temprow;
		V_I2(col) = tempcol;

		// Write a trace record
		sprintf(errline,"GetUnprotFieldRowCol for field %d row %d column %d\n",fieldNum, row, col);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
	}

	return rc;
}

short api3270::GetFieldRowCol(short fieldNum, VARIANT FAR* row, VARIANT FAR* col) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int		rc=0;
	char    errline[MAX_LINE + 50];
	short	temprow;
	short	tempcol;

	// Get the row and column attributes
	screen->CICS_3270_GetFieldLocByNum(fieldNum, &temprow, &tempcol, CICS_3270_FIELDS_ALL);

	if (temprow == -1)
	{
		// Error - invalid field number
		// indicate the error
		sprintf(errline, "03 GetFieldRowCol - no such field - %d\n", fieldNum);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to the caller with error
		rc = 3;
	}
 	else
	{
		// Clear the current data values
		VariantClear(row);
		VariantClear(col);

		// Set the variant data types to short
		V_VT(row) = VT_I2;
		V_VT(col) = VT_I2;

		// Set the data values in the variants
		V_I2(row) = temprow;
		V_I2(col) = tempcol;

		// Write a trace record
		sprintf(errline,"GetFieldRowCol for field %d row %d column %d\n",fieldNum, row, col);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
	}

	return rc;
}

short api3270::GetFieldLength(short row, short col, VARIANT FAR* length) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int		rc=0;	// return code
	CICS_3270_Field_t	*field;	// field characteristics
	char    errline[MAX_LINE + 50];

	//
	// Get the field referenced by the (row,column)
	//
	screen->CICS_3270_GetField(row, col, &field);

	// Check if there is a field here
	if (field->row == -1)
	{
		// Error - indicate no such field
		sprintf(errline, "02 GetFieldLength - no such field - (%d,%d)\n", row, col);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to the caller with error
		rc = 2;
	}
	else
	{
		// Clear the original variant data
		VariantClear(length);

		// Set the type to short
		V_VT(length) = VT_I2;

		// Get the result in the variant
		V_I2(length) = (short) field->length;
	}

	// release the data areas
	screen->CICS_3270_FreeField(field);

	return rc;
}

short api3270::keyin(LPCTSTR keyData) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char	errline[MAX_LINE + 50];
	short	rc;
	int		row;
	int		col;

	// Calculate the cursor position
	row = screen->cursorPosition/screen->numberColumns;
	col = screen->cursorPosition%screen->numberColumns;

	// Write a trace record
	sprintf(errline,"\nKeyin entered at (%d,%d) - %s\n",row,col,keyData);
	screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

	// Process the keystrokes
	rc = screen->CICS_3270_Keyin(keyData);

	return rc;
}

////////////////////////////////////////////////////////////////
//
// Routine to install a CICS terminal - must be called first
//
////////////////////////////////////////////////////////////////

short api3270::CreateTerminal(LPCTSTR regionName, LPCTSTR netName, LPCTSTR devType) 
{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int			rc=0;
	CICS_EpiDetails_t	termDet;
	char		errmsg[CICS_3270_MAX_EPI_MSG_LEN];
	char		errline[MAX_LINE + 50];

	// Write a trace record
	screen->CICS_3270_WriteTraceLine("\nCreateTerminal entered\n", CICS_3270_TRACE_MIN);

	// Check if the terminal is already installed
	if (TermIndex != CICS_EPI_TERM_INDEX_NONE)
	{
		// error - terminal already installed
		// log the error 
		strcpy(errline, "100: CreateTerminal - Terminal already installed\n");
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to caller with error
		return 100;
	}

	// Initialize the terminal characteristics and statistics
	InitializeTerminal();

	// Create an EPI Terminal and initialize it
   ////////////////////////////////////////////////////////////////
   //Install the logical terminal, returning detailed           
   //information about the installed terminal.                  
   ////////////////////////////////////////////////////////////////

   rc = CICS_EpiAddTerminal(
              NULL,
              (char *) regionName,
              (char *) netName,
              NULL,
              NULL,
              &termDet,
              &TermIndex);

   ////////////////////////////////////////////////////////////////
   // Check the return code from EpiAddTerminal.                 
   ////////////////////////////////////////////////////////////////

	if (rc != CICS_EPI_NORMAL)
	{
		// log the error 
		screen->CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
		sprintf(errline, "%d: CICS_EpiAddTerminal - %s\n", rc,errmsg);
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// return to caller with error
		return rc;
	}

   ////////////////////////////////////////////////////////////////
   // Capture the terminal name                                  
   ////////////////////////////////////////////////////////////////

	m_termname = termDet.NetName;

   ////////////////////////////////////////////////////////////////
   // If debug is enabled, write out the terminal details        
   ////////////////////////////////////////////////////////////////

   sprintf(errline,"Terminal Details for Terminal %s Index %d\n",
           termDet.NetName,TermIndex);
   screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

   sprintf(errline,"   Lines %d  Columns %d Max Data Size %d",
           termDet.NumLines, termDet.NumColumns, termDet.MaxData);

   if (termDet.Hilight == 1)
   {
      strcat(errline, "  Extended Highlighting");
   } // endif 

   if (termDet.Color == 1)
   {
      strcat(errline, "  Color");
   } // endif 

   strcat(errline, "\n");
   screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

   ////////////////////////////////////////////////////////////////
   // Initialize the 3270 data parsing module, specifying a      
   //  terminal model to match the installed model.              
   ////////////////////////////////////////////////////////////////

   switch (termDet.NumLines)
   {
       case 12:
          screen->CICS_3270_Initialize(1);
          break;

       case 24:
          screen->CICS_3270_Initialize(2);
          break;

       case 32:
          screen->CICS_3270_Initialize(3);
          break;

       case 43:
          screen->CICS_3270_Initialize(4);
          break;

       case 27:
          screen->CICS_3270_Initialize(5);
          break;

       default:
          screen->CICS_3270_Initialize(2);
          break;
   } // endswitch 

	return rc;
}

////////////////////////////////////////////////////////////////
//
// Routine to delete a previously installed terminal.
// This routine must be called if the terminal needs
// to be re-installed.  This might be necessary if the
// host system must be changed.
//
////////////////////////////////////////////////////////////////

short api3270::DeleteTerminal() 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short			rc = 0;
	cics_ushort_t	eventtype=0;

	// Check if the terminal is already installed
	if (TermIndex != CICS_EPI_TERM_INDEX_NONE)
	{
		// do the delete terminal
		rc = CICS_EpiDelTerminal(TermIndex);
		if (rc == CICS_EPI_NORMAL)
		{
			// wait for the end terminal event
			while((eventtype != CICS_EPI_EVENT_END_TERM) && (rc == 0))
			{
				rc = waitForEvent(&eventtype, CICS_EPI_WAIT);
			}
		}
	}
	else
	{
		// error - set return code to 100
		rc = 100;
	}

	TermIndex = CICS_EPI_TERM_INDEX_NONE;

	return rc;
}

////////////////////////////////////////////////////////////////
//
// Routine to handle an escape sequence for an 3270
// attention key.
//
////////////////////////////////////////////////////////////////

int api3270::processAID(char *cmdData)

{
   int          i=0;
   char         ch1;
   char         ch2;

   // translate the escape sequence characters to upper case 
   strupr(cmdData);

   ch1 = cmdData[1];
   ch2 = cmdData[2];

   switch (ch1)
   {
       case '0':
          if ((ch2 >= '1') && (ch2 <= '9'))
          {
             i = CICS_3270_AID_PF1 + (ch2 - '1');  // PF1-9 
          } // endif 
          break;

       case '1':
          switch (ch2)
          {
              case '0':
                 i = CICS_3270_AID_PF10;  // PF10 
                 break;

              case '1':
                 i = CICS_3270_AID_PF11;  // PF11 
                 break;

              case '2':
                 i = CICS_3270_AID_PF12;  // PF12 
                 break;

              case '3':
                 i = CICS_3270_AID_PF13;  // PF13 
                 break;

              case '4':
                 i = CICS_3270_AID_PF14;  // PF14 
                 break;

              case '5':
                 i = CICS_3270_AID_PF15;  // PF15 
                 break;

              case '6':
                 i = CICS_3270_AID_PF16;  // PF16 
                 break;

              case '7':
                 i = CICS_3270_AID_PF17;  // PF17 
                 break;

              case '8':
                 i = CICS_3270_AID_PF18;  // PF18 
                 break;

              case '9':
                 i = CICS_3270_AID_PF19;  // PF19 
                 break;

              default:
                 i = 0;
          } // endswitch 

          break;

       case '2':
          switch (ch2)
          {
              case '0':
                 i = CICS_3270_AID_PF20;  // PF20 
                 break;

              case '1':
                 i = CICS_3270_AID_PF21;  // PF21 
                 break;

              case '2':
                 i = CICS_3270_AID_PF22;  // PF22 
                 break;

              case '3':
                 i = CICS_3270_AID_PF23;  // PF23 
                 break;

              case '4':
                 i = CICS_3270_AID_PF24;  // PF24 
                 break;

              default:
                 i = 0;
          } // endswitch 

          break;

       case 'A':
          switch (ch2)
          {
              case '1':
                 i = CICS_3270_AID_PA1;  // PA1 
                 break;

              case '2':
                 i = CICS_3270_AID_PA2;  // PA2 
                 break;

              case '3':
                 i = CICS_3270_AID_PA3;  // PA3 
                 break;

              default:
                 i = 0;
          } // endswitch 

          break;

       case 'C':
          if (ch2 == 'L')
          {
             i = CICS_3270_AID_CLEAR;       // Clear 
          }
          else
          {
             if (ch2 == 'S')
             {
                i = CICS_3270_AID_PEN;      // Cursor Select 
             } // endif 
          } // endif 

          break;

       case 'E':
          if (ch2 == 'N')
          {
             i = CICS_3270_AID_ENTER;       // Enter 
          }

          break;

       case 'M':
          if (ch2 == 'E')
          {
             i = CICS_3270_AID_MSRE;       // End secure MSR 
          }

          break;

       case 'O':
          if (ch2 == 'P')
          {
             i = CICS_3270_AID_OPID;       // Operator ID 
          }

          break;

       case 'T':
          if (ch2 == 'R')
          {
             i = CICS_3270_AID_TRIG;       // Trigger Field 
          }

          break;

       default:
          i = 0;
   } // endswitch 

   return i;
}

char api3270::parseAid(LPCTSTR aidstr)

{
	char	aid='\0';		// initialize aid byte in case we don't recognize it
	char	pfnum='\0';		//work variable for PF key processing
	char	secondchar='\0';	//work variable for PF key processing

	// bypass any leading blanks
	while ((aidstr[0] == ' ') && (aidstr[0] != '\0'))
	{
		// skip leading blank
		aidstr++;
	}

	// Check for CLEAR key
	if (memcmp(aidstr, "CLEAR", 5) == 0)
	{
		aid = CICS_3270_AID_CLEAR;

		// Clear the input buffer
		screen->CICS_3270_ClearBuffer();
	}

	// Check for enter key
	if (memcmp(aidstr, "ENTER", 5) == 0)
	{
		aid = CICS_3270_AID_ENTER;
	}

	// Handle PF1-24 keys
	if (memcmp(aidstr, "PF", 2) == 0)
	{
		switch (aidstr[2])
		{
		case '0':
			pfnum = aidstr[3] - '0';
			break;
		case '1':
			if ((aidstr[3] >= '0') && (aidstr[3] <= '9'))
			{
				pfnum = (aidstr[3] - '0') + 10;
			}
			else
			{
				pfnum = aidstr[2] - '0';
			}
			break;
		case '2':
			if ((aidstr[3] >= '0') && (aidstr[3] <= '9'))
			{
				pfnum = (aidstr[3] - '0') + 20;
			}
			else
			{
				pfnum = aidstr[2] - '0';
			}
			break;
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			pfnum = aidstr[2] - '0';
			break;
		}
			
		switch (pfnum)
		{
		case 1:
			aid = CICS_3270_AID_PF1;
			break;
		
		case 2:
			aid = CICS_3270_AID_PF2;
			break;

		case 3:
			aid = CICS_3270_AID_PF3;
			break;

		case 4:
			aid = CICS_3270_AID_PF4;
			break;
		
		case 5:
			aid = CICS_3270_AID_PF5;
			break;

		case 6:
			aid = CICS_3270_AID_PF6;
			break;

		case 7:
			aid = CICS_3270_AID_PF7;
			break;
		
		case 8:
			aid = CICS_3270_AID_PF8;
			break;

		case 9:
			aid = CICS_3270_AID_PF9;
			break;

		case 10:
			aid = CICS_3270_AID_PF10;
			break;
		
		case 11:
			aid = CICS_3270_AID_PF11;
			break;

		case 12:
			aid = CICS_3270_AID_PF12;
			break;

		case 13:
			aid = CICS_3270_AID_PF13;
			break;
		
		case 14:
			aid = CICS_3270_AID_PF14;
			break;

		case 15:
			aid = CICS_3270_AID_PF15;
			break;

		case 16:
			aid = CICS_3270_AID_PF16;
			break;
		
		case 17:
			aid = CICS_3270_AID_PF17;
			break;

		case 18:
			aid = CICS_3270_AID_PF18;
			break;

		case 19:
			aid = CICS_3270_AID_PF19;
			break;
		
		case 20:
			aid = CICS_3270_AID_PF20;
			break;

		case 21:
			aid = CICS_3270_AID_PF21;
			break;

		case 22:
			aid = CICS_3270_AID_PF22;
			break;
		
		case 23:
			aid = CICS_3270_AID_PF23;
			break;

		case 24:
			aid = CICS_3270_AID_PF24;
			break;
		}
	}

	// Handle PA1-3 keys
	if (memcmp(aidstr, "PA", 2) == 0)
	{
		switch (aidstr[2])
		{
		case '0':
			aid = CICS_3270_AID_PA1;
			break;
		
		case '1':
			aid = CICS_3270_AID_PA2;
			break;

		case '2':
			aid = CICS_3270_AID_PA3;
			break;
		}
	}

	// Handle OPERID requests
	if (memcmp(aidstr, "OPERID", 6) == 0)
	{
		aid = CICS_3270_AID_OPID;
	}

	// Handle extended MSR requests
	if (memcmp(aidstr, "MSRE", 4) == 0)
	{
		aid = CICS_3270_AID_MSRE;
	}

	// Handle Trigger field requests
	if (memcmp(aidstr, "TRIGGER", 7) == 0)
	{
		aid = CICS_3270_AID_TRIG;
	}

	// Handle light pen or cursor select requests
	if (memcmp(aidstr, "PEN", 3) == 0)
	{
		aid = CICS_3270_AID_PEN;
	}

	// If aid byte not given, default to enter 
	if (aid == 0)
	{
		aid = CICS_3270_AID_ENTER;  // Enter key 
	} // endif 

	return aid;
}

////////////////////////////////////////////////////////////////
//
// Handle outbound 3270 data.  This function is called when data is received
// from CICS.  The internal model of the 3270 terminal is updated with the
// data.  It is also called when a transaction ends.
//
////////////////////////////////////////////////////////////////

 void api3270::HandleOutbound(u_byte_t *data, int length)

{
	 // Handle the data stream from the host
	screen->CICS_3270_Parse((char *)data, length);

	// update our screen parameter properties
	m_numberFields = (short) screen->CICS_3270_getNumberFields();
	m_numberUnprotFields = (short) screen->CICS_3270_getNumberUnprotFields();
	screen->CICS_3270_GetScreenParms(&m_rows, &m_columns);

    return;
}

////////////////////////////////////////////////////////////
//
// Routine to calculate the last response time
//
////////////////////////////////////////////////////////////

void api3270::saveLastResponse()

{
	int		newEventTime[MAX_SAVED_EVENTS + 1];

	// Remember the response time for each event
	memcpy(newEventTime + 1, lastEventTime, 4 * (MAX_SAVED_EVENTS - 1));
	newEventTime[0] = 0;
	memcpy(lastEventTime, newEventTime, 4 * MAX_SAVED_EVENTS);
}

////////////////////////////////////////////////////////////
//
// Routine to calculate the last response time
//
////////////////////////////////////////////////////////////

void api3270::calcLastResponse()

{
	SYSTEMTIME		curtime;
	FILETIME		filtime;
	LARGE_INTEGER	timeend;
	LARGE_INTEGER	differ;
	LARGE_INTEGER	const10000;

	// Get the current time
	GetSystemTime(&curtime);
	SystemTimeToFileTime(&curtime, &filtime);

	// Save the time in the work variable
	timeend.HighPart = filtime.dwHighDateTime;
	timeend.LowPart = filtime.dwLowDateTime;

	// Calculate the response time
	differ = LargeIntegerSubtract(timeend, timestart);

	// Convert to milliseconds
	const10000.HighPart = 0;
	const10000.LowPart = 10000;

	//differ = ExtendedMagicDivide(differ, const10000, 0);

	// Store the result in an instance variable
	m_lastResponse = (differ.LowPart / 10000);
	lastEventTime[0] = m_lastResponse;

	// Calculate the total Response time
	m_totalResponseTime += m_lastResponse;

	// Count the number of response times
	m_totalResponses++;
}

////////////////////////////////////////////////////////////
//
// Routine to process CICS EPI events
// It will return when either an end transaction or converse
// event is received, or an error occurs.  If no error is
// encountered, it will return a zero value.  If an error
// occurs, the error code will be returned.  The event
// type will be set to indicate the type of event
// which occurred.
//
////////////////////////////////////////////////////////////

int api3270::waitForEvent(cics_ushort_t * eventType, CICS_EpiWait_t waitType)

{
	int						rc;
	int						i;
	CICS_EpiEventData_t     Event;
	u_byte_t                dataBuffer[255*255];	// allocate a large buffer 
	char                    errmsg[CICS_3270_MAX_EPI_MSG_LEN];
	char					eventext[50];
	char                    errline[MAX_LINE + 50];
	char                    *data;
	int                     length;
	int						retfound = 1;	// indicate if we need to return
	char					newEventTable[MAX_SAVED_EVENTS];
	char					newEventData[4*MAX_SAVED_EVENTS];
	char					datalen[5];
	SYSTEMTIME				curtime;
	FILETIME				filtime;

	// Set the event type to zero
	*eventType = 0;

	// Enter the event loop
	while(1)
   {
   ////////////////////////////////////////////////////////////////
   // Wait for an event from CICS                                
   ////////////////////////////////////////////////////////////////

		Event.Data = dataBuffer;
		Event.Size = sizeof(dataBuffer);

		rc = CICS_EpiGetEvent(TermIndex,waitType,&Event);

		// Get error message if not normal return
		if (rc == CICS_EPI_NORMAL)
		{
			errmsg[0] = 0;
		}
		else
		{
			// Explain the return code
			screen->CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
		}

		// Write trace record
		eventmsg(Event.Event, eventext);
		sprintf(errline,"EpiGetEvent - event type %d %s - Return code = %d %s\n",
				Event.Event,eventext,rc,errmsg);
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

		// Check for errors
		if ((rc != CICS_EPI_NORMAL) && (rc != CICS_EPI_ERR_MORE_EVENTS) &&
			(rc != CICS_EPI_ERR_NO_EVENT))
		{
			// Write an error log message
			screen->CICS_3270_Write_Error_Log(errline);

			// Not sure what event if any was received
			eventType = 0;

			// Some kind of failure - return immediately
			return rc;
		}

   ////////////////////////////////////////////////////////////////
   // Handle the expected events.                                
   ////////////////////////////////////////////////////////////////

		// Get the event type
		*eventType = Event.Event;

		// Set the event type if we had a normal return code
		if ((rc == CICS_EPI_NORMAL) || (rc == CICS_EPI_ERR_MORE_EVENTS))
		{
			m_lastevent = Event.Event;

			// add the event to the event trace
			// get the previous events, dropping the last one,
			// insert the new one and copy back
			memcpy(newEventTable + 1, lastEvents, MAX_SAVED_EVENTS - 1);
			newEventTable[0] = (char) Event.Event;
			memcpy(lastEvents, newEventTable, MAX_SAVED_EVENTS);

			// similarly, we will shift the event data over one
			memcpy(newEventData + 4, lastEventData, 4 * (MAX_SAVED_EVENTS - 1));
			memset(newEventData, ' ', 4);	// clear the new entry to blanks
			memcpy(lastEventData, newEventData, 4 * MAX_SAVED_EVENTS);

			// Shift the event response times
			saveLastResponse();
		}

		switch(Event.Event)
		{
        //////////////////////////////////////////////////////////////////////////
        // The send event indicates that EXEC CICS SEND MAP ... was issued in   
        // the transaction.  The data stream received is parsed and modfies the 
        // internal representation of the terminal.                             
        //////////////////////////////////////////////////////////////////////////

		case CICS_EPI_EVENT_SEND:
			// Count the number of send events received
			m_sendCount++;

            HandleOutbound(Event.Data, Event.Size);

			// Remember the number of bytes receive
			m_bytesReceived += Event.Size;

			// save the length of the inbound data
			if (Event.Size > 9999)
			{
				strcpy(datalen, "9999");
			}
			else
			{
				sprintf(datalen,"%4.4d", Event.Size);
			}

			memcpy(lastEventData, datalen, 4);

			break;

        //////////////////////////////////////////////////////////////////////////
        // The converse event is expected when an EXEC CICS RECEIVE MAP ... is  
        // issued in the transaction.  This may include 3270 data stream not    
        // yet transmitted from the CICS region and hence outbound data must be 
        // handled first.  A CICS RECEIVE BUFFER is recognized and handled.     
        //////////////////////////////////////////////////////////////////////////

		case CICS_EPI_EVENT_CONVERSE:
			// Count the number of Converse Events Received
			m_convCount++;

			// Calculate the duration of the transaction
			calcLastResponse();

            if ((Event.Size == 1) &&
               (dataBuffer[0] == CICS_3270_CMD_RB))
            {
				// Count the read buffer request
				m_readBufferCount++;

				// handle a Read Buffer request 
				screen->CICS_3270_GetBuffer(39, &data, &length);
				rc = CICS_EpiReply(TermIndex,
								(u_byte_t *) data,
								(u_short_t) length);

				// check for errors 
				if (rc == CICS_EPI_NORMAL)
				{
					// Write a trace record
					sprintf(errline, "CICS_EpiReply to Read Buffer - %d bytes sent\n", length);
					screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

					// Remember the number of bytes sent
					m_bytesSent += length;
				}
				else
				{
					// log the error 
					screen->CICS_3270_GetEpiErrMsg(rc, TermIndex, errmsg);
					sprintf(errline, "Error %d CICS_EpiReply to Read Buffer - %s\n", rc,errmsg);
					screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
					screen->CICS_3270_Write_Error_Log(errline);

					// indicate we should return on the next 0 return code
					retfound = 0;
				}

				// Indicate this is really a read buffer
				lastEvents[0] = 99;
            }
            else
            {
				if (Event.Size > 0)
				{
					// Process any data sent by CICS
					// that has arrived with the request
					HandleOutbound(Event.Data, Event.Size);

					// Remember the number of bytes receive
					m_bytesReceived += Event.Size;

					// save the length of the inbound data
					if (Event.Size > 9999)
					{
						strcpy(datalen, "9999");
					}
					else
					{
						sprintf(datalen,"%4.4d", Event.Size);
					}

					memcpy(lastEventData, datalen, 4);
 				}

		  		// indicate we should return on the next 0 return code
 				retfound = 0;
			}

			break;

        ///////////////////////////////////////////////////////////////////
        // The end tran event is received when the transaction ends.     
        // The command file is then read, looking for the next           
        // transaction to execute.  If none is found, the startTrans     
        // will return a 1 and the terminal will be deleted.  This       
        // will trigger the end of the program.                          
        ///////////////////////////////////////////////////////////////////

		case CICS_EPI_EVENT_END_TRAN:
			// Increment the end transaction counter
			m_endTranCount++;

			// Calculate the duration of the transaction
			calcLastResponse();

            // no longer in a tran 
            m_transid.Empty();

            // remember the next tran id 
            // if specified on RETURN 
            m_nextransid = Event.TransId;

            sprintf(errline,"END TRAN event received - Next Transid is %s Response %d\n",m_nextransid,m_lastResponse);
            screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// remember the next transid in the trace data
			i = 0;
			while ((Event.TransId[i] != '\0') && (i < 4))
			{
				lastEventData[i] = Event.TransId[i];
				i++;
			}

			// indicate we should return on the next 0 return code
			retfound = 0;

			break;

        ///////////////////////////////////////////////////////////////////
        // This event is received when CICS has automatically started      
        // a transaction on the terminal.                                  
        ///////////////////////////////////////////////////////////////////

		case CICS_EPI_EVENT_START_ATI:
			// Increment the ATI transaction counter
			m_atiCount++;

			// Remember when the transaction was started
			// Get the current time
			GetSystemTime(&curtime);
			SystemTimeToFileTime(&curtime, &filtime);

			// Save the time in the work variable
			timestart.HighPart = filtime.dwHighDateTime;
			timestart.LowPart = filtime.dwLowDateTime;

			// A new transaction is now active
            sprintf(errline,"Start ATI event received for tran %s\n", Event.TransId);
            screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// Get the transaction id
			m_transid = Event.TransId;
			m_nextransid.Empty();

			// remember the next transid in the trace data
			i = 0;
			while ((Event.TransId[i] != '\0') && (i < 4))
			{
				lastEventData[i] = Event.TransId[i];
				i++;
			}

			// A new transaction is now active
			// indicate we should return on the next 0 return code
			retfound = 0;

			break;

        //////////////////////////////////////////////////////////////////
        //This event is received when the terminal has been uninstalled.         
		//////////////////////////////////////////////////////////////////

		case CICS_EPI_EVENT_END_TERM:
			// Enter the reason for the end terminal in the trace
			sprintf(errline,"End Terminal reason is %d\n",Event.EndReason);
            screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

			// indicate we should return on the next 0 return code
			retfound = 0;

			break;

		default:
			if (rc == CICS_EPI_NORMAL)
			{
				sprintf(errline,"Unknown Event received: %d\n", Event.Event);
				screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
			}

			// Really don't know what is going on, but better try returning
			retfound = 0;

			break;
      }

		// return when no more events pending
		if ((retfound == 0) && (rc != CICS_EPI_ERR_MORE_EVENTS))
		{
			return rc;
		}
   }

	return rc;
}

////////////////////////////////////////////////////////////////
//
// Routine to turn an EPI event type into a text message.
//

void api3270::eventmsg(short eventType, char* msg)

{
	switch (eventType)
	{
	case 0: strcpy(msg, "No event");
		    break;

	case 1: strcpy(msg, "Send");
			break;

	case 2: strcpy(msg, "Converse");
			break;

	case 3: strcpy(msg, "End Trans");
			break;

	case 4: strcpy(msg, "ATI");
			break;

	case 5: strcpy(msg, "End Term");
			break;

	default: sprintf(msg, "** Unknown event %d **",eventType);
			break;
	}
}

////////////////////////////////////////////////////////////////
//
// This routine will handle event processing.  If a send
// event is received, then the screen buffer will be updated.
// Control will be returned upon receipt of a converse,
// end transaction or end terminal event.
//
////////////////////////////////////////////////////////////////

short api3270::WaitEvent(VARIANT FAR* varEventType) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	rc;
	cics_ushort_t	event;

	// Write trace record
    screen->CICS_3270_WriteTraceLine("\nEntering WaitEvent\n", CICS_3270_TRACE_MIN);

	rc = waitForEvent(&event, CICS_EPI_WAIT);

	// Write trace record
    screen->CICS_3270_WriteTraceLine("Exiting WaitEvent\n", CICS_3270_TRACE_MIN);

	// Return the event code to the calling program
	if (rc == CICS_EPI_NORMAL)
	{
		// Clear the current variant value
		VariantClear(varEventType);

		// Set the variant type to short
		V_VT(varEventType) = VT_I2;

		// Get the event type data into the variant
		V_I2(varEventType) = event;
	}

	return rc;
}

/////////////////////////////////////////////////////////////
//
// This event is similar to WaitEvent, but in this case
// control will be returned immediately if no event is
// pending.
//
/////////////////////////////////////////////////////////////

short api3270::CheckEvent(VARIANT FAR* varEventType) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	rc;
	cics_ushort_t	event=0;

	// Write trace record
    screen->CICS_3270_WriteTraceLine("\nEntering CheckEvent\n", CICS_3270_TRACE_MIN);

	rc = waitForEvent(&event, CICS_EPI_NOWAIT);

	// Write trace record
    screen->CICS_3270_WriteTraceLine("Exiting CheckEvent\n", CICS_3270_TRACE_MIN);

	// Return the event code to the calling program
	if (rc == CICS_EPI_NORMAL)
	{
		// Clear the current variant value
		VariantClear(varEventType);

		// Set the variant type to short
		V_VT(varEventType) = VT_I2;

		// Get the event type data into the variant
		V_I2(varEventType) = event;
	}

	return rc;
}

///////////////////////////////////////////////////////
//    This routine handles ATI events
// It will enable for ATI events and check if a
// CICS transaction is started by the GetEvent.
//
//    A return of 99 indicates that a transaction
// is already running on the terminal.
// The program should wait for this transaction to
// end and issue a WaitEvent or CheckEvent.
// A return code of 16 should indicate that there
// are no EPI events queued for this terminal.
// A return code of 0 indicates that an ATI transaction
// has been started by CICS.  The program should now
// issue a WaitEvent or CheckEvent, and respond to any
// converse requests that it receives.  No further
// BeginTran, WaitATI or CheckATI requests can be
// made until an End Transaction event is received.
///////////////////////////////////////////////////////

short api3270::waitForATI(CICS_EpiWait_t waitType, cics_ushort_t* eventType, char* modname)

{
	int			rc=0;
	int			rc1=0;
	char		errline[MAX_LINE + 50];
	char		errmsg[CICS_3270_MAX_EPI_MSG_LEN];
	CICS_EpiATIState_t	ati_on=CICS_EPI_ATI_ON;
	CICS_EpiATIState_t	ati_off=CICS_EPI_ATI_HOLD;

	// Initialize the eventType
	*eventType = 0;

	// Write trace record
    sprintf(errline,"\nEntering %s\n",modname);
    screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

	// Check if a transaction is already running
	if (m_transid == "")
	{
		// Enable ATI requests
		rc = CICS_EpiATIState(TermIndex,&ati_on);

		if (rc == CICS_EPI_NORMAL)
		{
			// write a trace line
			screen->CICS_3270_WriteTraceLine("WaitforATI - ATI events enabled \n", CICS_3270_TRACE_MIN);

			// Wait for an ATI event
			rc = waitForEvent(eventType, waitType);

			// Check for errors
			if ((rc != CICS_EPI_NORMAL) && (rc != CICS_EPI_ERR_NO_EVENT))
			{
				// Error - Get Event failed
				sprintf(errline, "%d WaitforATI - Get Event failed\n", rc);

				// log and trace the error
				screen->CICS_3270_Write_Error_Log(errline);
				screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
			}

            // now turn ATI off again 
			// Use a different return code variable, so as to preserve
			// the return code from EpiGetEvent.  Otherwise, the 
			// application cannot tell if an ATI transaction was started
			// or not
            rc1 = CICS_EpiATIState(TermIndex,&ati_off);

            if (rc1 == CICS_EPI_NORMAL)
            {
				// write a trace line
				screen->CICS_3270_WriteTraceLine("WaitforATI - ATI events turned off \n", CICS_3270_TRACE_MIN);
            }
            else
            {
				// Build an error message
				screen->CICS_3270_GetEpiErrMsg(rc1, TermIndex, errmsg);
				sprintf(errline, "%d: Error turning CICS_EpiATIState off - %s\n",
						rc1, errmsg);
				// log and trace the error
				screen->CICS_3270_Write_Error_Log(errline);
				screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
            } // endif 
		}
		else
		{
			// Error - Enable ATI failed
			sprintf(errline, "WaitforATI - Enable ATI failed with rc = %d\n", rc);

			// log the error
			screen->CICS_3270_Write_Error_Log(errline);

			// write a trace record
			screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
		}
	}
	else
	{
		// Error - Transaction already running
		sprintf(errline, "99 WaitforATI - Transaction already running - %s", m_transid);

		// log the error
		screen->CICS_3270_Write_Error_Log(errline);

		// write a trace record
		screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);
		rc = 99;
 	}

	// Write trace record
	sprintf(errline,"Exiting %s - Return code = %d\n",modname,rc);
    screen->CICS_3270_WriteTraceLine(errline, CICS_3270_TRACE_MIN);

	return rc;
}

// This routine waits for an ATI request from the host.
// When an ATI request is started, the transaction 
// is executed.  This routine will return to the caller
// when the transaction ends or when the transaction 
// requests input data.

short api3270::WaitATI(VARIANT FAR* varEventType) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	rc;
	cics_ushort_t	event;

	// Execute the request
	rc = waitForATI(CICS_EPI_WAIT, &event, "WaitATI");

	// Return the event code to the calling program
	if (rc == CICS_EPI_NORMAL)
	{
		// Clear the current variant value
		VariantClear(varEventType);

		// Set the variant type to short
		V_VT(varEventType) = VT_I2;

		// Get the event type data into the variant
		V_I2(varEventType) = event;
	}

	return rc;
}

/////////////////////////////////////////////////////////////
//
// This event is similar to WaitATI, but in this case
// control will be returned immediately if no event is
// pending.
//
/////////////////////////////////////////////////////////////

short api3270::CheckATI(VARIANT FAR* varEventType) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short	rc;
	cics_ushort_t	event;

	// Execute the request
	rc = waitForATI(CICS_EPI_NOWAIT, &event, "CheckATI");

	// Return the event code to the calling program
	if (rc == CICS_EPI_NORMAL)
	{
		// Clear the current variant value
		VariantClear(varEventType);

		// Set the variant type to short
		V_VT(varEventType) = VT_I2;

		// Get the event type data into the variant
		V_I2(varEventType) = event;
	}

	return rc;
}

/////////////////////////////////////////////////////////////
//
// This routine will return a 16-bit check sum, calculated
// by adding the screen location of each attribute, with the
// first screen position being 1.  If there are not fields on
// the screen, the check sum will be zero.  The purpose
// of the check sum is to allow applications to quickly 
// distinguish between different screens.
//
/////////////////////////////////////////////////////////////

short api3270::GetScreenSum() 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int		screenSum = 0;

	screenSum = screen->CICS_3270_ScreenSum();

	return screenSum;
}

/////////////////////////////////////////////////////////////
//
// Routine to look up an EPI error and retun a textual
// error message.  This routine uses the EPIMSG.CPP
// program, which is separately compiled.
//
/////////////////////////////////////////////////////////////

BSTR api3270::EPIErrorMessage(short rc) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	CString strResult;
	char	msgtxt[512];

	// Get the error message
	screen->CICS_3270_GetEpiErrMsg(rc, TermIndex, msgtxt);

	// Get the value into a CString
	strResult = msgtxt;

	// return the value 
	return strResult.AllocSysString();
}

/////////////////////////////////////////////////////////////
//
// Routine to set trace on or off and set the trace file name
//
/////////////////////////////////////////////////////////////

short api3270::SetTrace(short reqtype, LPCTSTR tracefile) 
{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	short rc=0;

	switch (reqtype)
	{
	// Return the current trace status (1 is on, 0 is off)
	case 0: rc = screen->CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_INQ, tracefile);
			break;

	// Turn trace on, overwriting any existing trace file
	case 1: rc = screen->CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_ON, tracefile);
            screen->CICS_3270_WriteTraceLine("\n *** Trace set on ***\n", CICS_3270_TRACE_MIN);
			break;

	// Turn trace on, appending to an existing trace file
	case 2: rc = screen->CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_APPEND, tracefile);
            screen->CICS_3270_WriteTraceLine("\n *** Trace set on - append ***\n", CICS_3270_TRACE_MIN);
			break;

	// Turn trace off and close the trace file
	case 3: screen->CICS_3270_WriteTraceLine("\n *** Trace turned off ***\n", CICS_3270_TRACE_MIN);
			rc = screen->CICS_3270_ChgTrace(CICS_3270_TRACE_STATUS_OFF, tracefile);
			break;

	// Turn time stamps on
	case 4: rc = screen->CICS_3270_ChgTrace(CICS_3270_TRACE_DATE_ON, tracefile);
			break;

	// Turn time stamps off
	case 5: rc = screen->CICS_3270_ChgTrace(CICS_3270_TRACE_DATE_OFF, tracefile);
			break;
	}

	return rc;
}

/////////////////////////////////////////////////////////////
//
// Routine to set trace on or off and set the trace file name
//
/////////////////////////////////////////////////////////////

void api3270::SetTraceLevel(short tracelevel) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	screen->CICS_3270_SetTraceLevel(tracelevel);
}

void api3270::DumpFields() 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	screen->CICS_3270_DumpFields();
}

void api3270::WriteFields() 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	screen->CICS_3270_WriteFields();
}

void api3270::WriteTraceLine(LPCTSTR traceline, short tracelevel) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char fmtline[514];

	// print up to 512 bytes
	sprintf(fmtline, "%.512s\n",traceline);

	// Write the trace output
	screen->CICS_3270_WriteTraceLine(fmtline, tracelevel);
}

/////////////////////////////////////////////////////////////
//
// Routine to write a screen image to the trace file
//
/////////////////////////////////////////////////////////////

void api3270::TraceScreen(LPCTSTR protchar, LPCTSTR unprotchar) 

{
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char protc = ATTR_CHAR;
	char unprotc = UNPROT_CHAR;

	// check for a null character
	if (protchar[0] >= ' ')
	{
		protc = protchar[0];
	}

	// check for a null character
	if (unprotchar[0] >= ' ')
	{
		unprotc = unprotchar[0];
	}

	screen->CICS_3270_DumpBuffer(protc,unprotc);
}

/////////////////////////////////////////////////////////////
//
// This routine will return a string with a trace of the
// last MAX_SAVED_EVENTS events (default = 100);
//
/////////////////////////////////////////////////////////////

BSTR api3270::GetEventTrace() 

{
	CString strResult;
	// TODO: Add your dispatch handler code here
	// Make us threadsafe in using MFC
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	char	formatEvents[24 * MAX_SAVED_EVENTS + 1];
	int		i;
	int		respsecs;
	int		respmsec;
	char	eventype[9];
	char	tempevent[5];
	char	templine[50];

	// Initialize the result string
	formatEvents[0] = 0;
	
	for (i=0; i < MAX_SAVED_EVENTS; i++)
	{
		switch (lastEvents[i])
		{
		case 0:
			strcpy(eventype, "        ");
			break;
		case 1:
			strcpy(eventype, "SEND    ");
			break;
		case 2:
			strcpy(eventype, "CONVERSE");
			break;
		case 3:
			strcpy(eventype, "END TRAN");
			break;
		case 4:
			strcpy(eventype, "ATI STRT");
			break;
		case 5:
			strcpy(eventype, "END TERM");
			break;
		case 99:
			strcpy(eventype, "READBUFF");
			break;
		default:
			sprintf(eventype,"UNKN %3.3d",lastEvents[i]);
			break;
		}

		// Get the last event data
		memset(tempevent, 0, sizeof(tempevent));
		memcpy(tempevent, lastEventData + 4*i, sizeof(tempevent) - 1);
		while (strlen(tempevent) < 4)
		{
			strcat(tempevent, " ");
		}

		if (lastEventTime[i] == 0)
		{
			sprintf(templine, "%s %s          ",eventype,tempevent);
		}
		else
		{
			// Calculate the response time in milliseconds
			respmsec = lastEventTime[i]      ;
			respsecs = respmsec / 1000;		// Get the seconds part
			respmsec = respmsec % 1000;		// Get the milliseconds

			// Build the string
			sprintf(templine, "%s %s %d.%3.3d",eventype,tempevent,respsecs,respmsec);

			// pad to a minimum of 23 bytes
			while (strlen(templine) < 23)
			{
				strcat(templine, " ");
			}
		}

		// Append it to the end
		strcat(formatEvents, templine);
	}

	strResult = formatEvents;

	return strResult.AllocSysString();
}

/////////////////////////////////////////////////////////////
//
// This routine will get the list of the defined CICS servers
// from the CICS client.  The number of servers can be 
// obtained from the GetNumberServers interface, and the 
// name of a particular server can be obtained from the
// GetServerName interface.  This routine should only be
// called once, since it saves the table in memory.
//
/////////////////////////////////////////////////////////////

void initializeServerCount(short maxsys)

{
	// Get a pointer to the application object
	CEpi3270App* myApp = (CEpi3270App *) AfxGetApp();

	// First, allocate storage for up to MAX_CICS_SYSTEMS entries
	myApp->cicssystems = (CICS_EpiSystem_t *) malloc(sizeof(CICS_EpiSystem_t)*maxsys);

	// Clear the CICS system names
	 memset(myApp->cicssystems,0,sizeof(CICS_EpiSystem_t)*maxsys);

	// get the server names and descriptions
	myApp->numcicssystems = maxsys;
	int rc = CICS_EpiListSystems(NULL, &myApp->numcicssystems, myApp->cicssystems);
	if (rc != CICS_EPI_NORMAL)
	{
		// Didn't work, remember to clean up
		myApp->numcicssystems = 0;
		free(myApp->cicssystems);
	}
}

void initializeServers()

{
	// Get a pointer to the application object
	CEpi3270App* myApp = (CEpi3270App *) AfxGetApp();

	// Start with a reasonable amount of storage
	initializeServerCount(MAX_CICS_SYSTEMS);

	// Check if we exceeded our storage area
	if (myApp->numcicssystems > MAX_CICS_SYSTEMS)
	{
		free(myApp->cicssystems);
		// repeat the process with the right amount of storage
		initializeServerCount(myApp->numcicssystems);
	}
}

short api3270::ServerCount() 

{
	// TODO: Add your dispatch handler code here
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	// Get a pointer to the application object
	CEpi3270App* myApp = (CEpi3270App *) AfxGetApp();

	// Have we gotten the info from the ini file?
	if (myApp->numcicssystems == 0)
	{
		initializeServers();
	}

	return myApp->numcicssystems;
}

BSTR api3270::ServerName(short serverNumber) 

{
	CString strResult;
	// TODO: Add your dispatch handler code here
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	// Get a pointer to the application object
	CEpi3270App* myApp = (CEpi3270App *) AfxGetApp();

	// Have we gotten the info from the ini file?
	if (myApp->numcicssystems == 0)
	{
		initializeServers();
	}

	if ((serverNumber >= 0) && (serverNumber < myApp->numcicssystems))
	{
		// get the server name
		strResult = myApp->cicssystems[serverNumber].SystemName;
	}
	else
	{
		// Invalid server index
		// Return a empty string
		strResult.Empty();
	}

	return strResult.AllocSysString();
}

/////////////////
// End Of File //
/////////////////
