      *****************************************************************
      * VERSION:     1.0                                              *
      * AUTHOR:      JIM MACNAIR                                      *
      *                                                               *
      *    SIMPLE LINE MODE PROGRAM TO VALIDATE A USER ID AND         *
      *     PASSWORD AND SIGN ON A PARTICULAR TERMINAL.               *
      *                                                               *
      *    THE FORMAT OF THE TRANSACTION IS AS FOLLOWS:               *
      *                                                               *
      *     SGON USERID PASSWORD <NEW PASSWORD NEW PASSWORD>          *
      *                                                               *
      *    THE NEW PASSWORD MUST BE ENTERED TWICE.  THIS IS TO        *
      *     PREVENT THE USER OR PROGRAMMER FROM INADVERTANTLY         *
      *     CHANGING A PASSWORD BY HAVING SOME GARBAGE AT THE         *
      *     END OF THE INPUT LINE.                                    *
      *                                                               *
      *    THE COMMAND LINE PARAMETERS MUST BE SEPARATED BY ONE       *
      *     OR MORE BLANK CHARACTERS.                                 *
      *                                                               *
      *    DEPENDING ON THE TERMINAL INPUT, THERE MAY OR MAY NOT      *
      *    BE A SET BUFFER ADDRESS (SBA) SEQUENCE AT THE FRONT OF     *
      *    THE TERMINAL INPUT AREA.  IF FOUND, THE SBA SEQUENCE       *
      *    WILL BE IGNORED.                                           *
      *                                                               *
      *****************************************************************

       IDENTIFICATION DIVISION.
       PROGRAM-ID. SIGNONTS.
       ENVIRONMENT DIVISION.
       DATA DIVISION.
       WORKING-STORAGE SECTION.
       77  TERMLEN      PIC S9(4) COMP VALUE 0.
       77  MAXLEN       PIC S9(4) COMP VALUE 256.
       77  FLDCOUNT     PIC S9(4) VALUE 0.
       77  RESP         PIC S9(8) VALUE 0.
       77  RESP2        PIC S9(8) VALUE 0.
       77  TERMIDX      PIC S9(8) COMP.
       77  TRANSCODE    PIC X(4) VALUE SPACES.
       77  USERID       PIC X(8) VALUE SPACES.
       77  PWORD        PIC X(8) VALUE SPACES.
       77  NEWPW1       PIC X(8) VALUE SPACES.
       77  NEWPW2       PIC X(8) VALUE SPACES.
       77  REPLY        PIC X(100).
       01  TERMINPUT.
           02  TERMDATA  PIC X(256) VALUE SPACES.
           02  FILLER REDEFINES TERMDATA.
               03  TERMSBA PIC X.
               03  FILLER PIC X(2).
               03  TERMDAT2 PIC X(253).
       01  TERMIN  PIC X(256).
       01  BADREPLY.
           02   RETCODE  PIC 9(5).
           02   FILLER   PIC X VALUE SPACE.
           02   RETCODE2 PIC 9(5).
           02   FILLER   PIC X VALUE SPACE.
           02   BADMSG   PIC X(88) VALUE SPACES.
       01  FILLER.
      * MAKE BOTH BYTES A X'11'.
      * THIS WORKS ON PCS AND MAINFRAMES
           02 FILLER1 PIC S9(4) BINARY VALUE 4369.
           02 FILLER REDEFINES FILLER1.
              03 SBA  PIC X.
              03 FILLER  PIC X.

       PROCEDURE DIVISION.
      *
      * GET THE COMMAND LINE DATA INTO A BUFFER
      *
           EXEC CICS RECEIVE INTO(TERMINPUT) LENGTH(TERMLEN)
                     MAXLENGTH(MAXLEN) ASIS NOTRUNCATE
           END-EXEC.

      *
      * CHECK FOR AN SBA SEQUENCE AT THE FRONT OF THE
      * TERMINAL INPUT - IGNORE IT IF IT IS THERE
      *
           IF TERMSBA IS EQUAL TO SBA
      *
      *        GET RID OF THE SBA SEQUENCE AT THE FRONT
      *
               MOVE TERMDAT2 TO TERMIN
           ELSE
      *
      *        NO SBA SEQUENCE TO WORRY ABOUT
      *
               MOVE TERMDATA TO TERMIN
           END-IF

      *
      * PARSE THE COMMAND LINE
      *
           UNSTRING TERMIN DELIMITED BY ALL SPACES
             INTO TRANSCODE
                  USERID
                  PWORD
                  NEWPW1
                  NEWPW2
             TALLYING FLDCOUNT
           END-UNSTRING.

      *
      * CHECK THE USERID AND PASSWORD
      *
           IF FLDCOUNT IS GREATER THAN 3 THEN
             IF NEWPW1 IS EQUAL TO NEWPW2 THEN
      *        CHANGE THE PASSWORD
               EXEC CICS SIGNON USERID(USERID) PASSWORD(PWORD)
                    NEWPASSWORD(NEWPW1)
                    RESP(RESP) RESP2(RESP2)
               END-EXEC
             ELSE
               MOVE 1000 TO RESP2
               MOVE 1000 TO RESP
             END-IF
           ELSE
      *      NORMAL SIGNON - NO NEW PASSWORD
             EXEC CICS SIGNON USERID(USERID) PASSWORD(PWORD)
                  RESP(RESP) RESP2(RESP2)
             END-EXEC
           END-IF

      *
      *
      *
           IF RESP IS NOT EQUAL TO ZERO THEN
               MOVE RESP TO RETCODE
               MOVE RESP TO RETCODE2
               PERFORM 0200-CHECK-RESP2 THRU
                       0200-CHECK-RESP2-EXIT
               MOVE BADREPLY TO REPLY
           ELSE
               IF FLDCOUNT IS GREATER THAN 4
                 MOVE '00000 PASSWORD CHANGED' TO REPLY
               ELSE
                 MOVE '00000 TERMINAL SIGN ON COMPLETE' TO REPLY
               END-IF
           END-IF

      *
      * SEND THE RESULTS BACK TO THE TERMINAL
      *
           EXEC CICS SEND FROM(REPLY) LENGTH(LENGTH OF REPLY)
               ERASE
           END-EXEC.

      *
      * EXIT THE PROGRAM
      *
           EXEC CICS RETURN END-EXEC.
           GOBACK.

       0200-CHECK-RESP2.
           EVALUATE RESP2
             WHEN 1
               MOVE 'PASSWORD REQUIRED' TO BADMSG
             WHEN 2
               MOVE 'PASSWORD IS INCORRECT' TO BADMSG
             WHEN 3
               MOVE 'NEW PASSWORD REQUIRED' TO BADMSG
             WHEN 4
               MOVE 'NEW PASSWORD IS NOT ACCEPTABLE' TO BADMSG
             WHEN 8
               MOVE 'UNKNOWN USERID' TO BADMSG
             WHEN 9
               MOVE 'ALEADY SIGNED ON' TO BADMSG
             WHEN 10
               MOVE 'NON-TERMNAL TASK' TO BADMSG
             WHEN 12
               MOVE 'BAD RESPONSE FROM MODULE' TO BADMSG
             WHEN 13
               MOVE 'UNKNOWN ESMRESP OR NO ESM' TO BADMSG
             WHEN 14
               MOVE 'LANGUAGE NOT AVAILABLE' TO BADMSG
             WHEN 15
               MOVE 'MUST USE CRTE FOR ROUTING' TO BADMSG
             WHEN 16
               MOVE 'USERID NOT AUTH ON THIS TERMINAL' TO BADMSG
             WHEN 19
               MOVE 'USERID IS REVOKED' TO BADMSG
             WHEN 28
               MOVE 'INVALID NATIONAL LANGUAGE' TO BADMSG
             WHEN 29
               MOVE 'USER ALREADY SIGNED ON' TO BADMSG
             WHEN 30
               MOVE 'USERID ALL BLANK OR NULLS' TO BADMSG
             WHEN 200
               MOVE 'NOT ALLOWED IN DPL CALLED PROGRAM' TO BADMSG
             WHEN 1000
               MOVE 'NEW PASSWORDS DID NOT MATCH' TO BADMSG
             WHEN OTHER
               MOVE 'DID NOT RECOGNIZE RESP2' TO BADMSG
           END-EVALUATE.

       0200-CHECK-RESP2-EXIT.
           EXIT.
